from typing import *
import json
import re

from base_aux.base_enums.m0_enums import *
from base_aux.aux_argskwargs.m2_argskwargs_aux import *
from base_aux.base_source.m1_source import InitSource

from base_aux.aux_types.m0_types import TYPE__ELEMENTARY


# =====================================================================================================================
@final
class TextAux(InitSource):
    SOURCE: str

    def init_post(self) -> None | NoReturn:
        self.SOURCE = str(self.SOURCE)

    # EDIT ============================================================================================================
    def clear__spaces_all(self) -> str:
        """
        GOAL
        ----
        make a shortest string for like a str() from any container!
        assert str([1,2]) == "[1, 2]"
        assert func(str([1,2])) == "[1,2]"
        """
        self.SOURCE = re.sub(pattern=r" ", repl="", string=self.SOURCE)
        return self.SOURCE

    def clear__spaces_double(self) -> str:
        self.SOURCE = re.sub(pattern=r" {2,}", repl=" ", string=self.SOURCE)
        return self.SOURCE

    def clear__blank_lines(self) -> str:
        self.SOURCE = re.sub(pattern=r"^\s*\n", repl="", string=self.SOURCE, flags=re.MULTILINE)
        self.SOURCE = re.sub(pattern=r"\n\s*$", repl="", string=self.SOURCE, flags=re.MULTILINE)
        self.SOURCE = re.sub(pattern=r"^\s*$", repl="", string=self.SOURCE, flags=re.MULTILINE)  # not enough!
        return self.SOURCE

    def clear__cmts(self) -> str:
        """
        NOTE
        ----
        if oneline cmt - full line would be deleted!
        """
        self.SOURCE = re.sub(pattern=r"\s*\#.*$", repl="", string=self.SOURCE, flags=re.MULTILINE)
        return self.SOURCE

    # -----------------------------------------------------------------------------------------------------------------
    def strip__lines(self) -> str:
        self.lstrip__lines()
        self.rstrip__lines()
        return self.SOURCE

    def rstrip__lines(self) -> str:
        """
        GOAL
        ----
        keep indents! strip right!
            " line1 \n line2 " --> " line1\n line2"

        NOTE
        ----
        it can strip blank lines!
            " line1 \n \n  line2 " --> " line1\nline2"

        """
        self.SOURCE = re.sub(pattern=r"\s*$", repl="", string=self.SOURCE, flags=re.MULTILINE)
        return self.SOURCE

    def lstrip__lines(self) -> str:
        """
        NOTE
        ----
        less usefull as lstrip__lines
        but for the company)
        """
        self.SOURCE = re.sub(pattern=r"^\s*", repl="", string=self.SOURCE, flags=re.MULTILINE)
        return self.SOURCE

    # -----------------------------------------------------------------------------------------------------------------
    def sub__regexp(self, pat: str, new: str = "") -> str:
        """
        GOAL
        ----
        just a strait method
        """
        self.SOURCE = re.sub(pat, new, self.SOURCE)
        return self.SOURCE

    def sub__word(self, word_pat: str, new: str = "") -> str:
        """
        GOAL
        ----
        replace exact word(defined by pattern) in text.
        WORD means syntax word!

        SPECIALLY CREATED FOR
        ---------------------
        prepare_for_json_parsing
        """
        word_pat = r"\b" + word_pat + r"\b"
        self.SOURCE = re.sub(word_pat, new, self.SOURCE)
        return self.SOURCE

    def sub__words(self, rules: list[tuple[str, str]]) -> str:
        for work_pat, new in rules:
            self.sub__word(work_pat, new)
        return self.SOURCE

    # -----------------------------------------------------------------------------------------------------------------
    def prepare__json_loads(self) -> str:
        """
        GOAL
        ----
        replace pytonic values (usually created by str(Any)) before attempting to apply json.loads to get original python aux_types
        so it just same process as re.sub by one func for several values

        SPECIALLY CREATED FOR
        ---------------------
        try_convert_to_object
        """
        self.SOURCE = self.sub__words(
            rules = [
                (r"True", "true"),
                (r"False", "false"),
                (r"None", "null"),
            ]
        )
        # FIXME: apply work with int-keys in dicts!!! its difficalt to walk and edit result dict-objects in all tree!!!!
        self.SOURCE = re.sub("\'", "\"", self.SOURCE)
        return self.SOURCE

    def prepare__requirements(self) -> str:
        """
        GOAL
        ----
        replace pytonic values (usually created by str(Any)) before attempting to apply json.loads to get original python aux_types
        so it just same process as re.sub by one func for several values

        SPECIALLY CREATED FOR
        ---------------------
        try_convert_to_object
        """
        self.clear__cmts()
        self.clear__blank_lines()
        return self.SOURCE

    # =================================================================================================================
    def split_lines(self, skip_blanks: bool = None) -> list[str]:
        lines_all = self.SOURCE.splitlines()
        if skip_blanks:
            result_no_blanks = []
            for line in lines_all:
                if line:
                    result_no_blanks.append(line)
            return result_no_blanks

        else:
            return lines_all

    # =================================================================================================================
    def make__shortcut(
            self,
            maxlen: int = 15,
            where: Where3 = Where3.LAST,
            sub: str | None = "...",
    ) -> str:
        """
        MAIN IDEA-1=for SUB
        -------------------
        if sub is exists in result - means it was SHORTED!
        if not exists - was not shorted!
        """
        sub = sub or ""
        sub_len = len(sub)

        source = self.SOURCE
        source_len = len(source)

        if source_len > maxlen:
            if maxlen <= sub_len:
                return sub[0:maxlen]

            if where == Where3.FIRST:
                result = sub + source[-(maxlen - sub_len):]
            elif where == Where3.LAST:
                result = source[0:maxlen - sub_len] + sub
            elif where == Where3.MIDDLE:
                len_start = maxlen // 2 - sub_len // 2
                len_finish = maxlen - len_start - sub_len
                result = source[0:len_start] + sub + source[-len_finish:]
            else:
                result = source
            return result

        return source

    def make__shortcut_nosub(
            self,
            maxlen: int = 15,
            where: Where3 = Where3.LAST,
    ) -> str:
        """
        GOAL
        ----
        derivative-link for shortcut but no using subs!
        so it same as common slice
        """
        return self.make__shortcut(maxlen=maxlen, where=where, sub=None)

    # -----------------------------------------------------------------------------------------------------------------
    def make__object_try(self) -> TYPE__ELEMENTARY | str:
        """
        GOAL
        ----
        create an elementary object from text.
        or return source

        NOTE
        ----
        by now it works correct only with single elementary values like INT/FLOAT/BOOL/NONE
        for collections it may work but may not work correctly!!! so use it by your own risk and conscious choice!!
        """
        # FIXME: this is not work FULL and CORRECT!!!! need FIX!!!
        # PREPARE SOURCE ----------
        source_original = self.SOURCE

        # WORK --------------------
        try:
            result = self.prepare__json_loads()
            result = json.loads(result)
            return result
        except Exception as exx:
            print(f"{exx!r}")
            return source_original

    # =================================================================================================================
    def find__by_pats(self, patterns: list[str] | str) -> list[str]:
        """
        GOAL
        ----
        find all pattern values in text

        NOTE
        ----
        if pattern have group - return group value (as usual)
        """
        result = []
        patterns = ArgsKwargsAux(patterns).resolve_args()

        for pat in patterns:
            result_i = re.findall(pat, self.SOURCE)
            for value in result_i:
                value: str
                if value == "":
                    continue
                value = value.strip()
                if value not in result:
                    result.append(value)
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def requirements__get_list(self) -> list[str]:
        """
        GOAL
        ----
        get list of required modules (actually full lines stripped and commentsCleared)

        SPECIALLY CREATED FOR
        ---------------------
        setup.py install_requires
        """
        self.prepare__requirements()
        self.clear__blank_lines()
        self.strip__lines()
        result = self.split_lines()
        return result


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class PROJECT:
    # PROJECT ----------------------------------------------
    NAME_IMPORT: str = "requirements_checker"
    KEYWORDS: list[str] = [
        "check requirements", "raise/bool if no requirements",
        "check system requirements",
        "python packages/modules aux (upgrade/delete/version get)",
        "version parse", "version check", "version compare",
    ]

    # README -----------------------------------------------
    # add DOUBLE SPACE at the end of all lines! for correct representation in MD-viewers
    DESCRIPTION_SHORT: str = "check if requirements met"
    DESCRIPTION_LONG: str = """
designed for check requirements (systemOs) and raise/bool if no match
    """
    FEATURES: list[str] = [
        # "feat1",
        # ["feat2", "block1", "block2"],

        "check requirements (systemOs), raise/bool if no match",
        "create fuck(?)/source and is it for check for settings",
        ["[python PACKAGES/MODULES]", "upgrade", "delete", "version_get_installed", "check_installed)", "upgrade pip"],
        ["[VERSION]",
            "parse",
            "check",
            "compare",
        ],
    ]

    # HISTORY -----------------------------------------------
    VERSION: tuple[int, int, int] = (0, 2, 18)
    TODO: list[str] = [
        "add WARN_if__*/if_not__* (and use message in stderr)",
        "add check_file"
    ]
    FIXME: list[str] = [
        "sometimes modules have incorrect SHARE!!! maybe need check upgrade after installation!!! and show ERROR!",
        "FIX TESTS!"
    ]
    NEWS: list[str] = [
        "[Pkg] add check_prj_installed_latest +apply in upgrade_prj+share",
    ]


# =====================================================================================================================



from typing import *

import re

from base_aux.aux_types.m0_primitives import *
from base_aux.aux_argskwargs.m1_argskwargs import *
from base_aux.base_source.m2_source_kwargs import *
from base_aux.aux_types.m0_types import TYPE__VALID_VALIDATOR
from base_aux.aux_callable.m1_callable_aux import *
from base_aux.valid.m1_aux_valid_lg import *


# =====================================================================================================================
class _EqValidator:
    """
    MAIN IDEA
    ---------
    ALL WHAT PASSED INTO INIT WOULD PASS INTO VALIDATOR() AFTER FIRST ARG (OTHER)

    NOTE
    ----
    1/ preferably not use directly this object!
    USE DERIVATIVES!!! without validator passing

    2/ MAIN IDEA - NEVER RAISED!!! if any - return FALSE!!! if need - check manually!
    why so? - because i need smth to make a tests with final result of any source!
    dont mind reason!

    GOAL
    ----
    base object to make a validation by direct comparing with other object
    no raise

    USAGE
    -----
    PLACE ONLY IN FIRST PLACE!
    """
    VALIDATOR: TYPE__VALID_VALIDATOR

    V_ARGS: TYPE__ARGS_FINAL
    V_KWARGS: TYPE__KWARGS_FINAL

    REVERSE: bool = None

    OTHER_RAISED: bool = None
    OTHER_RESULT: Any | Exception = None

    def __init__(self, validator: TYPE__VALID_VALIDATOR = None, *v_args, reverse: bool = None, **v_kwargs) -> None:
        if validator is not None:
            self.VALIDATOR = validator

        if reverse is not None:
            self.REVERSE = reverse

        # super(ArgsKwargs, self).__init__(*v_args, **v_kwargs)
        self.V_ARGS = v_args
        self.V_KWARGS = v_kwargs

    def __str__(self):
        args = self.V_ARGS
        kwargs = self.V_KWARGS
        reverse = self.REVERSE
        return f"{self.__class__.__name__}({args=},{kwargs=},{reverse=})"

    def __eq__(self, other_draft) -> bool:
        return self.validate(other_draft)

    def __call__(self, other_draft: Any, *other_args, **other_kwargs) -> bool:
        """
        NOTE
        ----
        other_args/* - only for manual usage!
        typically used only other and only by direct eq(o1 == o2)
        """
        return self.validate(other_draft, *other_args, **other_kwargs)

    def __contains__(self, item) -> bool:
        return self.validate(item)

    def validate(self, other_draft: Any, *other_args, **other_kwargs) -> bool:
        """
        GOAL
        ----
        validate smth with special logic
        """
        # ------
        # TODO: decide use or not callable other??? = USE! it is really need to validate callable!!!
        try:
            self.OTHER_RESULT = CallableAux(other_draft).resolve_raise(*other_args, **other_kwargs)
            self.OTHER_RAISED = False
        except Exception as exx:
            self.OTHER_RAISED = True
            self.OTHER_RESULT = exx

        result = CallableAux(self.VALIDATOR).resolve_bool(self.OTHER_RESULT, *self.V_ARGS, **self.V_KWARGS)
        if self.REVERSE:
            result = not result
        return result

    def VALIDATOR(self, other_result, *v_args, **v_kwargs) -> bool | NoReturn:
        return NotImplemented


# ---------------------------------------------------------------------------------------------------------------------
class EqValid_Base(_EqValidator):
    def __init__(self, *v_args, reverse: bool = None, **v_kwargs):
        # print(v_args, v_kwargs)
        # super(ArgsKwargs, self).__init__(*v_args, **v_kwargs)     # not working!

        # super().__init__(*v_args, **v_kwargs)
        self.V_ARGS = v_args
        self.V_KWARGS = v_kwargs

        if reverse is not None:
            self.REVERSE = reverse


# =====================================================================================================================
@final
class EqValid_Variants(EqValid_Base):
    V_ARGS: TYPE__ARGS_FINAL
    V_KWARGS = KWARGS_FINAL__NOT_USED

    def VALIDATOR(self, other_result: Any, *variants: Any):
        if other_result in variants:
            return True
        else:
            return False


# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_VariantsStrLow(EqValid_Base):
    V_ARGS: tuple[str, ...]
    V_KWARGS = KWARGS_FINAL__NOT_USED

    def VALIDATOR(self, other_result: Any, *variants: Any):
        other_result = str(other_result).lower()
        variants = (str(var).lower() for var in variants)

        if other_result in variants:
            return True
        else:
            return False


# =====================================================================================================================
@final
class EqValid_Raise(EqValid_Base):
    """
    GOAL
    ----
    True - if Other object called with raised
    if other is exact final Exception without raising - it would return False!
    """
    V_ARGS = ARGS_FINAL__NOT_USED
    V_KWARGS = KWARGS_FINAL__NOT_USED

    def VALIDATOR(self, other_result, *v_args, **v_kwargs) -> bool:
        return self.OTHER_RAISED


# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_NotRaise(EqValid_Base):
    """
    GOAL
    ----
    True - if Other object called with raised
    if other is exact final Exception without raising - it would return False!
    """
    V_ARGS = ARGS_FINAL__NOT_USED
    V_KWARGS = KWARGS_FINAL__NOT_USED

    def VALIDATOR(self, other_result, *v_args, **v_kwargs) -> bool:
        return not self.OTHER_RAISED


# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_Exx(EqValid_Base):
    """
    GOAL
    ----
    True - if Other object is exact Exception or Exception()
    if raised - return False!!
    """
    V_ARGS = ARGS_FINAL__NOT_USED
    V_KWARGS = KWARGS_FINAL__NOT_USED

    def VALIDATOR(self, other_result, *v_args, **v_kwargs) -> bool | NoReturn:
        return not self.OTHER_RAISED and TypeAux(other_result).check__exception()


# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_ExxRaised(EqValid_Base):
    """
    GOAL
    ----
    True - if Other object is exact Exception or Exception() or Raised
    """
    V_ARGS = ARGS_FINAL__NOT_USED
    V_KWARGS = KWARGS_FINAL__NOT_USED

    def VALIDATOR(self, other_result, *v_args, **v_kwargs) -> bool | NoReturn:
        return self.OTHER_RAISED or TypeAux(other_result).check__exception()


# =====================================================================================================================
@final
class EqValid_LtGt(EqValid_Base):
    def VALIDATOR(self, other_result, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        return ValidAux(other_result).ltgt(low, high)


@final
class EqValid_LtGe(EqValid_Base):
    def VALIDATOR(self, other_result, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        return ValidAux(other_result).ltge(low, high)


@final
class EqValid_LeGt(EqValid_Base):
    def VALIDATOR(self, other_result, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        return ValidAux(other_result).legt(low, high)


@final
class EqValid_LeGe(EqValid_Base):
    def VALIDATOR(self, other_result, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        return ValidAux(other_result).lege(low, high)


# =====================================================================================================================
class EqValid_Regexp(EqValid_Base):
    V_ARGS: tuple[str, ...]
    V_KWARGS: TYPE__KWARGS_FINAL

    BOOL_COLLECT: BoolCollect = BoolCollect.ALL_TRUE

    def VALIDATOR(
            self,
            other_result,
            *regexps: str,
            ignorecase: bool = True,
            bool_collect: BoolCollect = None,
            match_link: Callable = re.fullmatch,
    ) -> bool | NoReturn:
        bool_collect = bool_collect or self.BOOL_COLLECT

        for pattern in regexps:
            result_i = match_link(pattern=str(pattern), string=str(other_result), flags=re.RegexFlag.IGNORECASE if ignorecase else 0)

            # CUMULATE --------
            if bool_collect == BoolCollect.ALL_TRUE:
                if not result_i:
                    return False
            elif bool_collect == BoolCollect.ANY_TRUE:
                if result_i:
                    return True
            elif bool_collect == BoolCollect.ALL_FALSE:
                if result_i:
                    return False
            elif bool_collect == BoolCollect.ANY_FALSE:
                if not result_i:
                    return True

        # FINAL ------------
        if bool_collect in [BoolCollect.ALL_TRUE, BoolCollect.ALL_FALSE]:
            return True
        else:
            return False


# ---------------------------------------------------------------------------------------------------------------------
@final
class EqValid_RegexpAllTrue(EqValid_Regexp):
    BOOL_COLLECT: BoolCollect = BoolCollect.ALL_TRUE


@final
class EqValid_RegexpAnyTrue(EqValid_Regexp):
    BOOL_COLLECT: BoolCollect = BoolCollect.ANY_TRUE


@final
class EqValid_RegexpAllFalse(EqValid_Regexp):
    BOOL_COLLECT: BoolCollect = BoolCollect.ALL_FALSE


@final
class EqValid_RegexpAnyFalse(EqValid_Regexp):
    BOOL_COLLECT: BoolCollect = BoolCollect.ANY_FALSE


# =====================================================================================================================
if __name__ == "__main__":
    pass


# =====================================================================================================================



from typing import *

from base_aux.base_source.m1_source import InitSource
from base_aux.aux_types.m1_type_aux import TypeAux

from base_aux.aux_argskwargs.m1_argskwargs import ArgsKwargs, TYPE__ARGS_DRAFT, TYPE__KWARGS_DRAFT, TYPE__ARGS_FINAL, TYPE__KWARGS_FINAL


# =====================================================================================================================
class ArgsKwargsAux(InitSource):
    SOURCE: TYPE__ARGS_DRAFT | TYPE__KWARGS_DRAFT

    def resolve_args(self) -> TYPE__ARGS_FINAL:     # REPLACING for args__ensure_tuple
        if isinstance(self.SOURCE, ArgsKwargs):
            return self.SOURCE.ARGS
        elif TypeAux(self.SOURCE).check__elementary_collection():
            return tuple(self.SOURCE)
        else:
            return (self.SOURCE,)

    def resolve_kwargs(self) -> TYPE__KWARGS_FINAL | NoReturn:
        if isinstance(self.SOURCE, ArgsKwargs):
            return self.SOURCE.KWARGS
        elif not self.SOURCE:
            return {}
        else:
            return dict(self.SOURCE)


# =====================================================================================================================



from typing import *
from base_aux.base_source.m1_source import InitSource


# =====================================================================================================================
TYPE__ITERABLE = Union[dict, list, tuple, set, Iterable]    # FIXME: add Iterable!!!

TYPE__ITER_PATH_KEY = Union[Any, int]
TYPE__ITER_PATH = tuple[TYPE__ITER_PATH_KEY, ...]


# =====================================================================================================================
@final
class IterAux(InitSource):
    """
    collect universal funcs which work with collections

    NOTE
    ----
    for access abilities passing with strings - resolve it by yourself

        assert self.victim("1/2", {1: 11, }) is None
        assert self.victim("1/2", {1: {2: 22}, }) == Explicit([1, 2, ])
        assert self.victim("1/2/1", {1: {2: [30, 31, 32]}, }) == Explicit([1, 2, 1])

        assert self.victim("hello", {"hello": [1]}) == Explicit(["hello", ])
        assert self.victim("hello/1", {"hello": [1]}) is None
        assert self.victim("hello/0", {"hello": [1]}) == Explicit(["hello", 0])

        assert self.victim("hello1/hello2", {"hello1": {"hello2": [1]}}) == Explicit(["hello1", "hello2"])
        assert self.victim("hello1/hello2/0", {"hello1": {"hello2": [1]}}) == Explicit(["hello1", "hello2", 0, ])
        assert self.victim("hello1/hello2/1", {"hello1": {"hello2": [1]}}) is None
    """
    SOURCE: TYPE__ITERABLE = dict
    # PATH: list[TYPE__ITER_PATH_KEY]   # todo: get back!!! to work with source! or make new class!

    # def init_post(self):
    #     self.PATH = []

    def item__get_original(self, item: Any) -> Any | None:
        """
        get FIRST original item from any collection by comparing str(expected).lower()==str(original).lower().

        # NOTE:
        # 1. NONE RESULT__VALUE - RESOLVED!!!
        # 2. SEVERAL VALUES - not used! by now it is just FIRST matched!
        #     several items? - it is not useful!!! returning first is most expected!
        #
        # USEFUL in case-insensitive systems (like terminals or serial devices) or object structured by prefix-names:
        # 1. get key in dict
        # 2. find attribute name in aux_types
        #
        # :param item:
        # :return: actual item from collection
        #     None - if VALUE is unreachable/notFind
        """
        for value in self.SOURCE:
            if value == item or str(value).lower() == str(item).lower():
                return value

    # -----------------------------------------------------------------------------------------------------------------
    def path__get_original(self, *path: TYPE__ITER_PATH_KEY) -> TYPE__ITER_PATH | None:
        """
        NOTES:
        1. path used as address KEY for dicts and as INDEX for other listed data
        2. separator is only simple SLASH '/'!

        :param path:
        :return:
            None - if path is unreachable/incorrect
            tuple[Any] - reachable path which could be used to get VALUE from data by chain data[i1][i2][i3]
        """

        # TODO: add attribute acceptable in path! and separate as Extruder?
        source = self.SOURCE
        if not path:
            return ()

        # work ----------------------------
        result = []
        for path_i in path:
            if isinstance(source, dict):
                address_original = IterAux(source).item__get_original(path_i)
                if not address_original:
                    return
                source = source[address_original]

            elif isinstance(source, set):
                msg = f"{source=},inconvenient type SET"
                raise TypeError(msg)

            else:
                # elif isinstance(source, (list, tuple)) or True:
                try:
                    address_original = int(path_i)
                    source = source[address_original]
                except:
                    return

            result.append(address_original)

        return tuple(result)

    # -----------------------------------------------------------------------------------------------------------------
    def value__get(self, *path: TYPE__ITER_PATH_KEY) -> Any | NoReturn:
        result = self.SOURCE
        path = self.path__get_original(*path)
        for path_i in path:
            result = result[path_i]
        return result

    def value__set(self, path: TYPE__ITER_PATH, value: Any) -> bool:
        """
        GOAL
        ----
        INLINE WORK!
        """
        source = self.SOURCE

        # work ----------------------------
        path = self.path__get_original(*path)
        try:
            length = len(path)
            for pos, path_i in enumerate(path, start=1):
                if pos == length:
                    source[path_i] = value
                else:
                    source = source[path_i]
        except:
            return False

        return True


# =====================================================================================================================



from typing import *

from base_aux.aux_values.m0_novalue import *
from base_aux.base_source.m2_source_kwargs import *


# =====================================================================================================================
@final
class ValidAux(InitSource):
    """
    Try to keep all validating funcs in separated place
    """
    # FIXME: do smth! before it was all static! and source at first place!
    SOURCE: Any = None

    def ltgt(self, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        result = True
        # low ----------
        if low is not None:
            result &= self.SOURCE > low
        if not result:
            return False
        # high ----------
        if high is not None:
            result &= self.SOURCE < high
        # final ----------
        return result

    def ltge(self, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        result = True
        # low ----------
        if low is not None:
            result &= self.SOURCE > low
        if not result:
            return False
        # high ----------
        if high is not None:
            result &= self.SOURCE <= high
        # final ----------
        return result

    def legt(self, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        result = True
        # low ----------
        if low is not None:
            result &= self.SOURCE >= low
        if not result:
            return False
        # high ----------
        if high is not None:
            result &= self.SOURCE < high
        # final ----------
        return result

    def lege(self, low: Any | None = None, high: Any | None = None) -> bool | NoReturn:
        result = True
        # low ----------
        if low is not None:
            result &= self.SOURCE >= low
        if not result:
            return False
        # high ----------
        if high is not None:
            result &= self.SOURCE <= high
        # final ----------
        return result


# =====================================================================================================================



from typing import *

from base_aux.aux_argskwargs.m2_argskwargs_aux import *


# =====================================================================================================================
class InitSourceKwArgs_Implicite(InitSource):
    """
    NOTE
    ----
    InitSourceKwArgs_Explicite is more useful!

    GOAL
    ----
    just to make inition source with KwArgs
    """
    ARGS: TYPE__ARGS_FINAL
    KWARGS: TYPE__KWARGS_FINAL

    def __init__(self, source: Any = None, *args, **kwargs) -> None:
        self.ARGS = args
        self.KWARGS = kwargs
        super().__init__(source)


# =====================================================================================================================
class InitSourceKwArgs_Explicite(InitSource):
    """
    MORE USEFUL THEN InitSourceKwArgs_Implicite

    GOAL
    ----

    FOR PYTESTAUX!
    """
    ARGS: TYPE__ARGS_FINAL
    KWARGS: TYPE__KWARGS_FINAL

    def __init__(self, source: Any = None, args: TYPE__ARGS_DRAFT = (), kwargs: TYPE__KWARGS_DRAFT = None, *args2, **kwargs2) -> None:
        self.ARGS = ArgsKwargsAux(args).resolve_args()
        self.KWARGS = ArgsKwargsAux(kwargs).resolve_kwargs()
        super().__init__(source, *args2, **kwargs2)


# =====================================================================================================================



from typing import *

from base_aux.base_exceptions import Exx__AnnotNotDefined
from base_aux.aux_types.m1_type_aux import TypeAux
from ..aux_types.m0_types import TYPES
from base_aux.base_source.m1_source import InitSource

from base_aux.aux_argskwargs.m1_argskwargs import *

from .m1_attr1_aux import AttrAux


# =====================================================================================================================
@final
class AnnotsAux(InitSource):
    """
    GOAL
    ----
    work with all __annotations__
        from all nested classes
        in correct order

    RULES
    -----
    4. nesting available with correct order!
        class ClsFirst(BreederStrStack):
            atr1: int
            atr3: int = None

        class ClsLast(BreederStrStack):
            atr2: int = None
            atr4: int

        for key, value in ClsLast.annotations__get_nested().items():
            print(f"{key}:{value}")

        # atr1:<class 'int'>
        # atr3:<class 'int'>
        # atr2:<class 'int'>
        # atr4:<class 'int'>
    """
    # -----------------------------------------------------------------------------------------------------------------
    def get_not_defined(self) -> list[str]:
        """
        GOAL
        ----
        return list of not defined annotations

        SPECIALLY CREATED FOR
        ---------------------
        annot__check_all_defined
        """
        result = []
        nested = self.dump__dict_types()
        for key in nested:
            if not AttrAux(self.SOURCE).anycase__check_exists(key):
                result.append(key)
        return result

    # =================================================================================================================
    def check_all_defined(self) -> bool:
        """
        GOAL
        ----
        check if all annotated aux_attr have value!
        """
        return not self.get_not_defined()

    def check_all_defined_or_raise(self) -> None | NoReturn:
        """
        GOAL
        ----
        check if all annotated aux_attr have value!
        """
        not_defined = self.get_not_defined()
        if not_defined:
            dict_type = self.dump__dict_types()
            msg = f"[CRITICAL]{not_defined=} in {dict_type}"
            raise Exx__AnnotNotDefined(msg)

    # =================================================================================================================
    def dump__dict_types(self) -> dict[str, type[Any]]:
        """
        GOAL
        ----
        get all annotations in correct order (nesting available)!

        RETURN
        ------
        keys - all attr names (defined and not)
        values - Types!!! not instances!!!
        """

        result = {}
        for cls in self.iter_mro():
            _result_i = dict(cls.__annotations__)
            _result_i.update(result)
            result = _result_i
        return result

    def dump__dict_values(self) -> TYPE__KWARGS_FINAL:
        """
        GOAL
        ----
        get dict with only existed values! no raise if value not exists!
        """
        result = {}
        for key in self.dump__dict_types():
            if hasattr(self.SOURCE, key):
                result.update({key: getattr(self.SOURCE, key)})
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def dump__pretty_str(self) -> str:
        """just a pretty string for debugging or research.
        """
        result = f"{self.SOURCE.__class__.__name__}(Annotations):"
        annots = self.dump__dict_values()
        if annots:
            for key, value in annots.items():
                result += f"\n\t{key}={value}"
        else:
            result += f"\nEmpty=Empty"

        print(result)
        return result

    # =================================================================================================================
    def iter_mro(self) -> Iterable[type]:
        """
        GOAL
        ----
        iter only important user classes from mro
        """
        yield from TypeAux(self.SOURCE).iter_mro_user(AnnotsBase)

    def iter_names(self) -> Iterable[str]:
        """
        iter all (with not existed)
        """
        yield from self.dump__dict_types()

    def iter_values(self) -> Iterable[Any]:
        """
        only existed
        """
        yield from self.dump__dict_values().values()

    # =================================================================================================================
    def values__set_none__existed(self) -> None:
        """
        GOAL
        ----
        set None for all annotated aux_attr! only existed!
        """
        for name in self.iter_names():
            if hasattr(self.SOURCE, name):
                setattr(self.SOURCE, name, None)

    def values__set_none__all(self) -> None:
        """
        GOAL
        ----
        set None for all annotated aux_attr! even not existed!
        """
        for name in self.iter_names():
            setattr(self.SOURCE, name, None)

    def values__delete(self) -> None:
        """
        GOAL
        ----
        delattr all annotated aux_attr!
        """
        for name in self.iter_names():
            if hasattr(self.SOURCE, name):
                delattr(self.SOURCE, name)

    def values__reinit_by_types(self, not_existed: bool = None) -> None:
        """
        GOAL
        ----
        delattr all annotated aux_attr!
        """
        for name, value in self.dump__dict_types().items():
            if not_existed and hasattr(self.SOURCE, name):
                continue

            value = TypeAux(value).type__init_value__default()
            setattr(self.SOURCE, name, value)

    # -----------------------------------------------------------------------------------------------------------------
    def names__delete(self, *names: str) -> None:
        """
        ATTENTION
        ---------
        its not good idea!!! dont use it! yoг will change real CLASS parameters|

        GOAL
        ----
        del names from annots! if existed
        """
        if not names:
            names = self.iter_names()

        for name in names:
            for cls in self.iter_mro():
                if name in cls.__annotations__:
                    cls.__annotations__.pop(name)

    def names__add(self, *names: str) -> None:
        """
        ATTENTION
        ---------
        its not good idea!!! dont use it! yoг will change real CLASS parameters|

        GOAL
        ----
        add names into annots!
        """
        cls = TypeAux(self).get__class()

        for name in names:
            annots = cls.__annotations__
            if name not in annots:
                cls.__annotations__.update({name: Any})


# =====================================================================================================================
class AnnotsBase:
    # -----------------------------------------------------------------------------------------------------------------
    def __getattr__(self, name) -> Any | NoReturn:
        return AttrAux(self).anycase__getattr(name)

    def __getitem__(self, name) -> Any | NoReturn:
        return AttrAux(self).anycase__getattr(name)


# =====================================================================================================================



from typing import *

import re
from base_aux.aux_types.m0_primitives import *
from base_aux.aux_argskwargs.m1_argskwargs import *
from base_aux.base_source.m2_source_kwargs import *
from base_aux.aux_types.m0_types import TYPE__VALID_VALIDATOR
from base_aux.aux_callable.m1_callable_aux import *
from base_aux.aux_eq.m2_eq_validator import *


# =====================================================================================================================
@final
class EqValidChain(EqValid_Base):
    V_ARGS: tuple[EqValid_Base, ...]
    V_KWARGS: TYPE__KWARGS_FINAL    # TODO: add params for AllTrue/Any*/False*

    def validate(self, other_draft: Any) -> bool:
        other_final = other_draft

        for eq_i in self.V_ARGS:
            if eq_i != other_final:
                return False

            other_final = eq_i.OTHER_RESULT

        return True


# =====================================================================================================================
if __name__ == "__main__":
    pass


# =====================================================================================================================



from typing import *


# =====================================================================================================================
@final
class AttrsInitByKwArgs:
    """
    GOAL
    ----
    create object with aux_attr as dict model / template

    SAME AS
    -------
    AttrAux.load_by_dict but more simple!

    SPECIALLY CREATED FOR
    ---------------------
    best way seems in test issues for base_aux.args

    NOTE
    ----
    used only acceptable items! no raise!
    """
    def __init__(self, *args, **kwargs) -> None:
        self.__init_args(*args)
        self.__init_kwargs(**kwargs)

    def __init_kwargs(self, **kwargs) -> None:
        for name, value in kwargs.items():
            try:
                setattr(self, name, value)
            except:
                pass

    def __init_args(self, *args) -> None:
        kwargs = dict.fromkeys(args)
        self.__init_kwargs(**kwargs)


# =====================================================================================================================



import pathlib
from typing import *

from base_aux.aux_types.m2_info import *
from base_aux.base_resolver.m1_resolver import *


# =====================================================================================================================
class ResolveFilePath(Resolver):
    """
    GOAL
    ----
    resolve filepath by any variant

    SPECIALLY CREATED FOR
    ---------------------
    base_aux.files
    """
    NAME: str = None
    EXTLAST: str = None
    DIRPATH: pathlib.Path = None

    # as properties -------
    NAMEEXT: str
    FILEPATH: pathlib.Path

    @property
    def NAMEEXT(self) -> str:
        result = ""
        if self.NAME is not None:
            result += f"{self.NAME}"
        if self.EXTLAST is not None:
            result += f".{self.EXTLAST}"
        return result

    @property
    def FILEPATH(self) -> pathlib.Path:
        return self.DIRPATH.joinpath(self.NAMEEXT)

    def __init__(
            self,
            name: str = None,
            extlast: str = None,
            dirpath: str | pathlib.Path = None,
            nameext: str = None,

            filepath: str | pathlib.Path = None,
    ):
        """
        NOTE
        ----
        you can use "filepath" as base/default and others (name/extlast/...) for overwrite some of them base parts
        """

        if filepath is not None:
            filepath = pathlib.Path(filepath)
            self.DIRPATH = filepath.parent
            self.NAME = filepath.stem
            self.EXTLAST = filepath.suffix.rsplit(".", 1)[-1]

        if dirpath is not None:
            self.DIRPATH = pathlib.Path(dirpath)
        if self.DIRPATH is None and dirpath is None:
            self.DIRPATH = pathlib.Path().cwd()

        if nameext is not None:
            name_ext: list[str] = nameext.rsplit(".", 1)
            if len(name_ext) == 2:  # DOT exists!
                name, extlast = name_ext
                if extlast:
                    self.EXTLAST = extlast
                if name:
                    self.NAME = name
            else:
                self.NAME = nameext

        # most important! overwrite previous set!
        if name is not None:
            self.NAME = name
        if extlast is not None:
            self.EXTLAST = extlast

    # -----------------------------------------------------------------------------------------------------------------
    def resolve(self) -> pathlib.Path:
        return self.FILEPATH


# =====================================================================================================================
if __name__ == '__main__':
    obj = pathlib.Path("hello.")
    ObjectInfo(obj).print()


# =====================================================================================================================



# =====================================================================================================================
# USE COMMON TYPES

_std = [
# base ----
BaseException,
Exception,

# USER ----
UserWarning,
BytesWarning,
Warning,
DeprecationWarning,
PendingDeprecationWarning,

AssertionError,

NotImplemented,
NotImplementedError,

# std ------
TypeError,
NameError,
ValueError,
AttributeError,

SyntaxWarning,
SyntaxError,

# imports -----
ImportError,
ImportWarning,
ModuleNotFoundError,

WindowsError,
IOError,
OSError,
EnvironmentError,
SystemError,
PermissionError,
ChildProcessError,
MemoryError,

KeyError,
KeyboardInterrupt,

FileExistsError,
FileNotFoundError,

ConnectionError,
ConnectionAbortedError,
ConnectionResetError,
ConnectionRefusedError,
TimeoutError,
EOFError,
BufferError,

LookupError,
IndexError,

EncodingWarning,

UnicodeWarning,
UnicodeDecodeError,
UnicodeEncodeError,
UnicodeTranslateError,

ZeroDivisionError,
UnboundLocalError,
TabError,
SystemExit,

BrokenPipeError,

StopIteration,
StopAsyncIteration,
RuntimeWarning,
ResourceWarning,
ReferenceError,
RecursionError,
ProcessLookupError,
OverflowError,
RuntimeError,
NotADirectoryError,
IsADirectoryError,
InterruptedError,
IndentationError,
GeneratorExit,
FutureWarning,
FloatingPointError,
ExceptionGroup,
BlockingIOError,
BaseExceptionGroup,

ArithmeticError,

# REAL VALUE = NOT AN EXCEPTION!!!
NotImplemented,      # NotImplemented = None # (!) real value is 'NotImplemented'
]

# =====================================================================================================================
class Exx__AnnotNotDefined(Exception):
    """Exception in case of not defined attribute in instance
    """


class Exx__NumberArithm_NoName(Exception):
    pass


class Exx__GetattrPrefix(Exception):
    pass


class Exx__GetattrPrefix_RaiseIf(Exx__GetattrPrefix):
    pass


class Exx__ValueNotParsed(Exception):
    pass


class Exx__ValueUnitsIncompatible(Exception):
    pass


class Exx__IndexOverlayed(Exception):
    pass


class Exx__IndexNotSet(Exception):
    pass


class Exx__ItemNotExists(Exception):
    """
    not exists INDEX (out of range) or NAME not in defined values
    """
    pass


class Exx__StartOuterNONE_UsedInStackByRecreation(Exception):
    """
    in stack it will be recreate automatically! so dont use in pure single BreederStrSeries!
    """
    pass


class Exx__BreederObjectList_GroupsNotGenerated(Exception):
    pass


class Exx__BreederObjectList_GroupNotExists(Exception):
    pass


class Exx__BreederObjectList_ObjCantAccessIndex(Exception):
    pass


# =====================================================================================================================
class Exx__Valid(Exception):
    pass


class Exx__ValueNotValidated(Exx__Valid):
    pass


# =====================================================================================================================



from typing import *
from typing import Union, Any, Callable, NoReturn

from base_aux.aux_values.m0_novalue import TYPE__NOVALUE


# =====================================================================================================================
class _Cls:
    def meth(self):
        pass


# =====================================================================================================================
@final
class TYPES:
    """
    GOAL
    ----
    collect all types variants
    """

    # SINGLE ---------------------------
    NONE: type = type(None)
    FUNCTION: type = type(lambda: True)
    METHOD: type = type(_Cls().meth)

    # COLLECTIONS ---------------------------
    ELEMENTARY_SINGLE: tuple[type, ...] = (
        type(None),
        bool,
        int, float,
        str, bytes,
    )
    ELEMENTARY_COLLECTION: tuple[type, ...] = (
        tuple, list,
        set, dict,
    )
    ELEMENTARY: tuple[type, ...] = (
        *ELEMENTARY_SINGLE,
        *ELEMENTARY_COLLECTION,
    )


# =====================================================================================================================
TYPE__ELEMENTARY = Union[*TYPES.ELEMENTARY]


# =====================================================================================================================
TYPE__VALID_EXX = Union[Exception, type[Exception]]
TYPE__VALID_RESULT = Union[
    Any,
    TYPE__VALID_EXX,  # as main idea! instead of raise
]
TYPE__VALID_SOURCE_BOOL = Union[
    Any,                                # fixme: hide? does it need? for results like []/{}/()/0/"" think KEEP! it mean you must know that its expecting boolComparing in further logic!
    bool,                               # as main idea! as already final generic
    Callable[[...], bool | Any | NoReturn],   # as main idea! to get final generic
    # TYPE__VALID_EXX,
    TYPE__NOVALUE
]
TYPE__VALID_RESULT_BOOL = Union[
    # this is when you need get only bool! raise - as False!
    bool,  # as main idea! instead of raise/exx
]
TYPE__VALID_RESULT_BOOL__EXX = Union[
    bool,
    TYPE__VALID_EXX,
]
TYPE__VALID_VALIDATOR = Union[
    Any,    # generic final instance as expecting value - direct comparison OR comparison instance like Valid!
    # Type,   # Class as validator like Valid? fixme
    TYPE__VALID_EXX,  # direct comparison
    Callable[[Any, ...], bool | NoReturn]     # func with first param for validating source
]


# =====================================================================================================================
